/* File: functions.c */


#include "functions.h"


void interrupt_handler(int signo)
{
    /*an ekteleitai to parakolouthoumeno ektelesimo*/
    if(tr_exe_pid != 0)
    {
        /*tou proothei to sima kai stamataei na to parakolouthei*/
        ptrace(PTRACE_DETACH, tr_exe_pid, NULL, signo);
        /*epanaferei sto 0 ti metavliti eto pid tou*/
        tr_exe_pid = 0;

        /*an yparxei energi diergasia gia anakateuthinsi tou stdin*/
        if (spids->stdin_pid != 0)
        {
            kill(spids->stdin_pid, signo);
            waitpid(spids->stdin_pid, NULL, 0);
            /*an to arxeio gia tin anakateuthinsi stdin den exei kleisei*/
            if(fin != NULL)
            {
                fclose(fin);
            }             
        }
        /*an yparxei energei diergasia gia anakateuthinsi tou stdout*/
        if (spids->stdout_pid != 0)
        {
            kill(spids->stdout_pid, signo);
            waitpid(spids->stdout_pid, NULL, 0);
            /*an to arxeio gia tin anakateuthinsi stdout den exei kleisei*/
            if(fout != NULL)
            {
                fclose(fout);
            }            
        }
        /*an yparxei energi diergasia gia anakateuthinsi tou stderr*/        
        if (spids->stderr_pid != 0)
        {
            kill(spids->stderr_pid, signo);
            waitpid(spids->stderr_pid, NULL, 0);
            /*an to arxeio gia tin anakateuthinsi stderr den exei kleisei*/
            if(ferr != NULL)
            {
                fclose(ferr);
            }            
        }            
    } 
    else
    {
        /*epanaferei tin default symperifora ton simaton kai stelnei to sima ston eauto tou*/
        act.sa_handler =  SIG_DFL;
        sigaction(signo, &act , NULL); 
        raise(signo);        
    }
}


void print_welcome()
{
    printf("\nWelcome to picodb! Type your desired command or type help to see the available\n"
           "commands and what they do, then hit Enter.\n\n");      
}


void print_wait_input()
{
    printf("->");
}


void print_wrong_command()
{
    printf("Wrong command! Type help to see the available commands.\n");
}


void print_help()
{
    printf("trace <category>\t\tApplying for monitoring system calls of the"
                " <category> class.\n"
                "\t\t\t\tThe parameter <category> can be:\n"
                "\t\t\t\t- process-control: system calls related\n"
                "\t\t\t\tto processes(execve, fork, wait4, kill).\n"
                "\t\t\t\t- file-management: system calls related\n"
                "\t\t\t\tto file-management(open, close, read, write).\n"
                "\t\t\t\t- all: all the categories above.\n"
                "\t\t\t\t- none: none of the categories above.\n"
        "\n"
        "redirect <stream> <filename>\trequesting redirection of one of the three standard\n"
                "\t\t\t\tinput/output streams of the program will be executed from/to a file.\n"
                "\t\t\t\tThe parameter <filename> is the name of the file.\n"
                "\t\t\t\tThe <stream> can be:\n"
                "\t\t\t\t- stdin: requesting redirection of the input of the\n"
                "\t\t\t\tprogram that will be executed from the file <filename>.\n"
                "\t\t\t\t- stdout: requesting redirection of the stdout output of the\n"
                "\t\t\t\tprogram that will be executed to the file <filename>.\n"
                "\t\t\t\t- stderr: requesting redirection of the stderr output of the\n"
                "\t\t\t\tprogram that will be executed to the file <filename>.\n"
        "\n"
        "blocking-mode <mode>\t\tSets the blocking-mode parameter in the state <mode>.\n"
                "\t\t\t\tThe parameter <mode> can be:\n"
                "\t\t\t\t- on: to freeze the execution of the traced program\n"
                "\t\t\t\twhenever one of the system calls being monitored is called.\n"
                "\t\t\t\t- off: the program runs without interruption.\n"
        "\n"
        "limit-trace <number>\t\tSets a limit on how many system calls each class will be monitored.\n"
                "\t\t\t\tIf the limit-trace is set, after <number> calls of system\n"
                "\t\t\t\tcalls of each class, all other calls will be ignored.\n"
                "\t\t\t\tIf the limit-trace is not set or a number less than 1 is given, \n"
                "\t\t\t\tthe picodb will monitor all calls of the system calls that their categories\n"
                "\t\t\t\t where defined through trace command. \n"
        "\n"
        "go <args>\t\t\tRuns the program that was given to picodb as argument\n"
                "\t\t\t\tusing the <args> parameter to pass any arguments to it,\n"
                "\t\t\t\tand monitors it according to the commands that were given.\n"    
        "\n"
        "quit\t\t\t\tTerminates the picodb.\n"
        "\n"
        "help\t\t\t\tPrints all the available commands that picodb gives\n"
             "\t\t\t\tto the user with a small description.\n\n");    
}


void trace_command(char *input, char *temp, int *tr_cat_pc, int *tr_cat_fm)
{
    if(sscanf(input, "%*s %s", temp) == EOF)
    {
        print_wrong_command();
        print_wait_input();
    }
    else if(!strcmp(temp, "process-control"))
    {
        *tr_cat_pc = 1;
        printf("Monitoring proccess control system calls.\n");
        print_wait_input();
    }
    else if(!strcmp(temp, "file-management"))
    {
        *tr_cat_fm = 1;
        printf("Monitoring file management system calls.\n");
        print_wait_input();
    }        
    else if(!strcmp(temp, "all"))
    {
        *tr_cat_pc = 1;
        *tr_cat_fm = 1;
        printf("Monitoring proccess control and file management system calls.\n");
        print_wait_input();
    }   
    else if(!strcmp(temp, "none"))
    {
        *tr_cat_pc = 0;
        *tr_cat_fm = 0;
        printf("Monitoring none of the system calls.\n");
        print_wait_input();
    }             
    else
    {
        print_wrong_command();
        print_wait_input();
    }    
}


void redirect_command(char *input, char *temp, 
                      int *rd_stdin, int *rd_stdout, int *rd_stderr,
                      char *rd_infile, char *rd_outfile, char *rd_errfile)
{
    if(sscanf(input, "%*s %s", temp) == EOF)
    {
        print_wrong_command();
        print_wait_input();
    }
    else if(!strcmp(temp, "stdin"))
    {                
        if(sscanf(input, "%*s %*s %s", temp) == EOF)
        {
            print_wrong_command();
            print_wait_input();
        }
        else
        {
            FILE *ftemp;
            ftemp = fopen(temp, "r");
            if(ftemp != NULL)
            {
                *rd_stdin = 1;
                strcpy(rd_infile, temp);
                fclose(ftemp);
                printf("The stdin will be redirected from the file: %s\n", rd_infile);
                print_wait_input();
            }
            else
            {
                printf("The file %s does not exist.\n", temp);
                print_wait_input();
            }
        }                
    }
    else if(!strcmp(temp, "stdout"))
    {                
        if(sscanf(input, "%*s %*s %s", temp) == EOF)
        {
            print_wrong_command();
            print_wait_input();
        }
        else
        {
            *rd_stdout = 1;
            strcpy(rd_outfile, temp);
            printf("The stdout will be redirected to the file: %s\n", rd_outfile);
            print_wait_input();
        }                
    }
    else if(!strcmp(temp, "stderr"))
    {                
        if(sscanf(input, "%*s %*s %s", temp) == EOF)
        {
            print_wrong_command();
            print_wait_input();
        }
        else
        {
            *rd_stderr = 1;
            strcpy(rd_errfile, temp);
            printf("The stderr will be redirected to the file: %s\n", rd_errfile);
            print_wait_input();
        }                
    }
    else if(!strcmp(temp, "clear"))
    {   
        *rd_stdin = 0;
        *rd_stdout = 0;
        *rd_stderr = 0;
        printf("The redirections were cleared.\n");
        print_wait_input();            
    }    
    else
    {
        print_wrong_command();
        print_wait_input();
    }    
}


void block_command(char *input, char *temp, int *blk_mode)
{
    if(sscanf(input, "%*s %s", temp) == EOF)
    {
        print_wrong_command();
        print_wait_input();
    }
    else if(!strcmp(temp, "on"))
    {                
        *blk_mode = 1;
        printf("Blocking mode activated.\n");
        print_wait_input();
    }
    else if(!strcmp(temp, "off"))
    {                
        *blk_mode = 0;
        printf("Blocking mode deactivated.\n");
        print_wait_input();
    }            
    else
    {
        print_wrong_command();
        print_wait_input();
    }     
}


void limit_command(char *input, char *temp, int *lmt_trace, int *lmt_number)
{
    if(sscanf(input, "%*s %s", temp) == EOF)
    {
        print_wrong_command();
        print_wait_input();
    }
    else
    {                
        *lmt_number = atoi(temp);
        if(*lmt_number > 0)
        {
            *lmt_trace = 1;
            printf("A limit of %d was set to the number of the traced system calls.\n", *lmt_number);
            print_wait_input();
        }
        else
        {
            *lmt_trace = 0;
            printf("No limit to the number of the traced system calls.\n");
            print_wait_input();
        }
    }    
}


int go_pc_categories(int orig_eax, int lmt_trace, int lmt_number, int blk_mode, pid_t pid,  
                      int *tcpc_tmp, int *insyscall, int *cur_num_pc, int *calls)
{
    if (orig_eax == SYS_execve)
    {
        if (*insyscall == 0)
        {
            if(lmt_trace)
            {
                (*cur_num_pc)++;
                if(*cur_num_pc == lmt_number)
                {
                    *tcpc_tmp = 0;
                    *insyscall = 0;
                }
            }

            printf("The child made an execve system call\n");
            *insyscall = 1;

            if(block(blk_mode))
            {
                kill(pid, SIGKILL);
                waitpid(pid, NULL, 0);
                return 1;
            }                                

            if(*calls == 1)
            {
                *insyscall = 0;
            }
            (*calls)++;
        }
        else
        {
            if(*calls%2 == 0)
            {
                *insyscall = 0;  
                (*calls)++;
            }
            (*calls)++;
        }
    }
    if (orig_eax == SYS_clone)
    {
        if (*insyscall == 0)
        {    
            if(lmt_trace)
            {
                (*cur_num_pc)++;
                if(*cur_num_pc == lmt_number)
                {
                    *tcpc_tmp = 0;
                    *insyscall = 0;
                }
            }

            printf("The child made a fork system call\n");
            *insyscall = 1;

            if(block(blk_mode))
            {
                kill(pid, SIGKILL);
                waitpid(pid, NULL, 0);
                return 1;
            }                                                                                                 
        }
        else
        {
            *insyscall = 0;
        }
    }
    if (orig_eax == SYS_wait4)
    {
        if (*insyscall == 0)
        {                                
            if(lmt_trace)
            {
                (*cur_num_pc)++;
                if(*cur_num_pc == lmt_number)
                {
                    *tcpc_tmp = 0;
                    *insyscall = 0;
                }
            }

            printf("The child made a wait system call\n");
            *insyscall = 1;

            if(block(blk_mode))
            {
                kill(pid, SIGKILL);
                waitpid(pid, NULL, 0);
                return 1;
            }                                 
        }
        else
        {                                
            *insyscall = 0;
        }
    }    
    if (orig_eax == SYS_kill)
    {
        if (*insyscall == 0)
        {                            
            if(lmt_trace)
            {
                (*cur_num_pc)++;
                if(*cur_num_pc == lmt_number)
                {
                    *tcpc_tmp = 0;
                    *insyscall = 0;
                }
            }

            printf("The child made a kill system call\n");
            *insyscall = 1;

            if(block(blk_mode))
            {
                kill(pid, SIGKILL);
                waitpid(pid, NULL, 0);
                return 1;
            }

            (*calls)++;                                
        }
        else
        { 
            if(*calls%2 == 0)
            {
                *insyscall = 0; 
                (*calls)++;
            }
            (*calls)++;
        }
    }    
    
    return 0;
}


int go_fm_categories(int orig_eax, int lmt_trace, int lmt_number, int blk_mode, pid_t pid,  
                      int *tcfm_tmp, int *insyscall, int *cur_num_fm)
{
    if (orig_eax == SYS_write)
    {
        if (*insyscall == 0)
        {                              
            if(lmt_trace)
            {
                (*cur_num_fm)++;
                if(*cur_num_fm == lmt_number)
                {
                    *tcfm_tmp = 0;
                    *insyscall = 0;
                }
            }

            printf("The child made a write system call\n");
            fflush(stdout);
            *insyscall = 1;

            if(block(blk_mode))
            {
                kill(pid, SIGKILL);
                waitpid(pid, NULL, 0);
                return 1;
            }
        }
        else
        { 
            *insyscall = 0;
        }
    }
    if (orig_eax == SYS_read)
    {
        if (*insyscall == 0)
        {                               
            if(lmt_trace)
            {
                (*cur_num_fm)++;
                if(*cur_num_fm == lmt_number)
                {
                    *tcfm_tmp = 0;
                    *insyscall = 0;
                }
            }

            printf("The child made a read system call\n");
            *insyscall = 1;

            if(block(blk_mode))
            {
                kill(pid, SIGKILL);
                waitpid(pid, NULL, 0);
                return 1;
            }
        }
        else
        { 
            *insyscall = 0;
        }
    }
    if (orig_eax == SYS_close)
    {
        if (*insyscall == 0)
        {                               
            if(lmt_trace)
            {
                (*cur_num_fm)++;
                if(*cur_num_fm == lmt_number)
                {
                    *tcfm_tmp = 0;
                    *insyscall = 0;
                }
            }

            printf("The child made a close system call\n");
            *insyscall = 1;

            if(block(blk_mode))
            {
                kill(pid, SIGKILL);
                waitpid(pid, NULL, 0);
                return 1;
            }
        }
        else
        { 
            *insyscall = 0;  
        }
    }    
    if (orig_eax == SYS_open)
    {
        if (*insyscall == 0)
        {                               
             if(lmt_trace)
            {
                (*cur_num_fm)++;
                if(*cur_num_fm == lmt_number)
                {
                    *tcfm_tmp = 0;
                    *insyscall = 0;
                }
            }

            printf("The child made an open system call\n");
            *insyscall = 1;

            if(block(blk_mode))
            {
                kill(pid, SIGKILL);
                waitpid(pid, NULL, 0);
                return 1;
            }
        }
        else
        { 
            *insyscall = 0;  
        }
    }    
    
    return 0;
}

void go_get_args(char* input, char* tr_exe_args[], char* tr_exe)
{
    int i;
    char *token = strtok(input, " \n");
    for(i = 0; token; i++)
    {
        if(i == 0)
        {
            tr_exe_args[i] = tr_exe; 
        }
        else
        {
            tr_exe_args[i] = token;                    
        }
        token = strtok(NULL, " \n");
    } 
    tr_exe_args[i] = NULL;    
}

int block(int blk_mode)
{
    int stop = 0;
    
    if(blk_mode)
    {
        char in[10];        

        printf("Type y to continue or n to stop, then hit Enter: ");
        scanf("%s", in);
        while (1)
        {
            if(!strcmp(in, "y"))
            {
                break;
            }
            else if(!strcmp(in, "n"))
            {
                stop = 1;
                break;
            }
            else
            {
                printf("You must type y or n: ");
                scanf("%s", in);
            }
        }
    }
    
    return stop;
}


void init_shmemseg()
{
    shmseg_id = shmget(IPC_PRIVATE, sizeof(streams_pids*), S_IRUSR | S_IWUSR);
    spids = (streams_pids*) shmat(shmseg_id, NULL, 0);    
}


void det_shmemseg()
{
    shmdt((void*)spids);
}


void rm_shmemseg()
{
    shmctl(shmseg_id, IPC_RMID, NULL);    
}
